/*    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2013-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package dd.util;

import java.util.Comparator;

import org.apache.commons.math3.distribution.NormalDistribution;

/** A contingency table. */
public class ContingencyTable {
	// initializer block
	static {
		setConfLevel(0.95);
	}
	/** CT Specification */
	public CTSpecs specs;
	/** Context. */
	public FrequentItemset frset;
	/** a value. */
	public int a;
	/** b value. */
	public int b;
	/** c value. */
	public int c;
	/** d value. */
	public int d;

	/** Constructor. */
	public ContingencyTable(CTSpecs specs) {
		this.specs = specs;
	}
	
	/** Constructor. */
	public ContingencyTable(int a, int b, int c, int d) {
		this.a = a;
		this.b = b;
		this.c = c;
		this.d = d;
	}
	
	/** Constructor. */
	public ContingencyTable() {
	}

	/** Measure list. */
	public enum Measure {RD, RR, RC, OR, ED, ER, EC, SD};

	/** Get measure code. */
	public static Measure getMeasureCode(String m) {
		switch(m) {
			case "RD": return Measure.RD;
			case "RR": return Measure.RR;
			case "RC": return Measure.RC;
			case "OR": return Measure.OR;
			case "ED": return Measure.ED;
			case "ER": return Measure.ER;
			case "EC": return Measure.EC;
			case "SD": return Measure.SD;
			default: throw new RuntimeException("Measure " + m + " unknown");
		}
	}

	/** Get measure type. */
	public static boolean isSDlike(Measure m) {
		if(m==Measure.SD)
			return true;
		return false;
	}
	
	/** Get measure value. */
	public double getMeasure(Measure m) {
		switch(m) {
			case RD: return rd();
			case RR: return rr();
			case RC: return rc();
			case OR: return or();
			case ED: return ed();
			case ER: return er();
			case EC: return ec();
			case SD: return sd();
			default: throw new RuntimeException("Measure " + m + " unknown");
		}
	}

	/** n1 value. */
	public int n1()  {
		return a+b;
	}

	/** n2 value. */
	public int n2()  {
		return c+d;
	}

	/** m1 value. */
	public int m1()  {
		return a+c;
	}

	/** m2 value. */
	public int m2()  {
		return b+d;
	}

	/** n value. */
	public int n()  {
		return a+b+c+d;
	}

	/** p1 value. */
	public double p1()  {
		int n1 = a+b;
		if( n1 > 0)
			return (double)a/n1;
		return specs.avgNegative;
	}

	/** p2 value. */
	public double p2()  {
		int n2 = c+d;
		if( n2 > 0)
			return (double)c/n2;
		return specs.avgNegative;
	}

	/** p value. */
	public double p()  {
		return (double)(a+c)/(a+b+c+d);
	}

	/** RD measure. */
	public double rd()  {
		return p1()-p2();
	}
	
	/** RR measure. */
	public double rr()  {
		double p2 = p2();
		if(p2==0) p2 = (double)1/(d+1);
		return p1()/p2;
	}
	
	/** RC measure. */
	public double rc()  {
		double p2 = p2();
		if(p2==1) p2 = (double)c/(c+1);
		return (1-p1())/(1-p2);
	}
	
	/** OR measure. */
	public double or()  {
		double p1 = p1();
		if(p1==1) p1 = (double)a/(a+1);
		double p2 = p2();
		if(p2==0) p2 = (double)1/(d+1);
		return p1/(1-p1)*(1-p2)/p2;
	}
	
	/** ED measure. */
	public double ed()  {
		return p1()-p();
	}

	/** SD measure. */
	public double sd()  {
		int m2 = m2();
		if( m2 > 0)
			return n1()/(double)n() - (double)b/m2;
		return n1()/(double)n() - specs.avgProtected; 
	}

	/** ER measure. */
	public double er()  {
		double p = p();
		if(p==0) p = (double)1/(n()+1);
		return p1()/p;
	}

	/** EC measure. */
	public double ec()  {
		double p = p();
		if(p==0) p = (double)1/(n()+1);
		return (1-p1())/p;
	}

	/** Maximum variational distance. */
	public double tval(boolean sdType) {
		double maxdiff = -1;
		if( sdType ) {
			int m2 = m2();
			if( m2 > 0)
				maxdiff =  Math.abs((double)b/m2 - specs.avgProtected);
			int m1 = m1();
			if( m1 > 0) {
				double d = Math.abs((double)a/m1 - specs.avgProtected);
				if( d > maxdiff )
					maxdiff = d;
			}
			return maxdiff;
		}
		if( n1() > 0) 
			maxdiff =  Math.abs(p1() - specs.avgNegative);
		if( n2() > 0) {
			double d = Math.abs(p2() - specs.avgNegative);
			if( d > maxdiff )
				maxdiff = d;
		}
		if( maxdiff == - 1 ) 
			throw new RuntimeException("ContingencyTable.tval() on empty values: a = " + a + " b = " + b + " c = " + c + " d = " +d + " avgNegative = " + specs.avgNegative);
		return maxdiff;
	}

	public String toString() {
		return "("+a+", "+b+", "+c+", "+d+")";
	}
	
	/** Interface for comparators. */
	public interface CompM {
		public double measure(ContingencyTable ct);
	}
	
	/** Comparator on a parametric measure. */
	public static class CompMeasure implements Comparator<ContingencyTable> {
		private CompM c;
		public CompMeasure(CompM c) {
			this.c = c;			
		}
	    public int compare(ContingencyTable ct1, ContingencyTable ct2) {
	    	// order lexicographic on (measure, support, itemset)
	    	double measure1 = c.measure(ct1);
	    	double measure2 = c.measure(ct2);
	    	if(measure1<measure2)
	    		return -1;
	    	if(measure1>measure2)
	    		return 1;
	    	int supp1 = ct1.n();
	    	int supp2 = ct2.n();
	    	if(supp1<supp2)
	    		return -1;
	    	if(supp1>supp2)
	    		return 1;
	    	int cmpPD = ct1.specs.pdItemset.compareTo(ct2.specs.pdItemset);
	    	if( cmpPD != 0 )
	    		return cmpPD;
	    	return ct1.frset.compareTo(ct2.frset);
	    }
	}

	/** Comparator on the RR measure. */
	public static class CompRR implements CompM {
		public double measure(ContingencyTable ct) {
			return ct.rr();
		}
	}
	
	/** Comparator on the RD measure. */
	public static class CompRD implements CompM {
		public double measure(ContingencyTable ct) {
			return ct.rd();
		}
	}
	
	/** Confidence level in statistical tests. */
	public static double CL;
	/** Z in statistical tests. */
	public static double Z;
	/** Z^2 in statistical tests. */
	public static double Zsq;
	/** Z^2/2 in statistical tests. */
	public static double Zsq2;

	public static void setConfLevel(double cl)
	{
		CL = cl;
		NormalDistribution n = new NormalDistribution(); 
		Z = n.cumulativeProbability((1+cl)/2);
		Zsq = Z*Z;
		Zsq2 = Zsq/2;
/*
		System.out.println("Z = " + Z);
		System.out.println("Zsq = " + Zsq);
		System.out.println("Zsq2 = " + Zsq2);
*/
	}
	
	/** RR lower bound cache. */
	private static CTCache<Double> cacheRRlb = new CTCache<Double>();
	/** RD lower bound cache. */
	private static CTCache<Double> cacheRDlb = new CTCache<Double>();
	/** Use caching in confidence interval calculations. */
	public static boolean useCache = true; 
	
	/** Confidence intervals for RR. */
	public double ciRRlb() {
	    if(useCache) {
			Double cache = cacheRRlb.get(a,b,c,d);
			if(cache != null)
				return cache;
	    }
		double dx1 = a + Zsq2;
		double dn1 = n1() + Zsq;
		double dx2 = c + Zsq2;
		double dn2 = n2() + Zsq;
		double p1 = dx1/dn1;
		double p2 = dx2/dn2;
	    double rhat = p1/p2;
	    double delta = Math.exp(Z * Math.sqrt(1/dx1-1/dn1+1/dx2-1/dn2));
	    double rrLB = rhat/delta;
	    if(useCache)
	    	cacheRRlb.put(a,b,c,d,rrLB);
	    return rrLB;
	}

	/** Confidence intervals for RD. */
	public double ciRDlb() {
	    if(useCache) {
			Double cache = cacheRDlb.get(a,b,c,d);
			if(cache != null)
				return cache;
	    }
		double dx1 = a + Zsq2;
		double dn1 = n1() + Zsq;
		double dx2 = c + Zsq2;
		double dn2 = n2() + Zsq;
		double p1 = dx1/dn1;
		double p2 = dx2/dn2;
	    double rhat = p1-p2;
	    double delta = Z * Math.sqrt(p1*(1-p1)/dn1 + p2*(1-p2)/dn2);
	    double rdLB =  rhat - delta;
	    if(useCache)
	    	cacheRDlb.put(a,b,c,d,rdLB);
	    return rdLB;
	}

	/** Comparator on the RR CI lower bound measure. */
	public static class CompRRlb implements CompM {
		public double measure(ContingencyTable ct) {
			return ct.ciRRlb();
		}
	}
	
	/** Comparator on the RD CI lower bound measure. */
	public static class CompRDlb implements CompM {
		public double measure(ContingencyTable ct) {
			return ct.ciRDlb();
		}
	}
}
