/*    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2013-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package dd.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.TreeSet;

import dd.util.ContingencyTable.Measure;

import rst.pp.fimi.CSV2FIMI;
import rst.pp.fimi.Decode;
import rst.table.BitTable;
import rst.table.TableSubset;
import rst.util.data.Bitmap;
import rst.util.IOUtil;
import rst.util.SubsetGen;

/** Table for Discrimination Discovery analyses. */
public class DDTable {

	/** File system separator. */
	public static final String fs = "/";

	/** Working directory. */
	public String wDir;
	/** Sub-directory for data. */
	public String wDirData;
	/** Dataset base-name. */
	public String dataset;
	/** Decode itemsets. */
	public Decode dec;
	/** Dataset in BitTable format. */
	public BitTable table;
	/** Number of rows. */
	public int nRows;

	/** PD items as String. */
	public String pdItemString;
	/** PD items. */
	public ArrayList<Integer> pdItems;
    /** Powerset of PD Itemsets specifications. */
    public CTSpecs []specs;
	/** Negative decision item as String. */
	public String decItemString;
	/** Negative decision item. */
	public int decisionItem;

	/** Minimum support for contingency table extraction. */
	public int support;
    /** Extract closed itemsets or frequent itemsets. */
    public boolean closedItemset = true;
	   
	/** Reader of closed itemsets. */
    private BufferedReader closedin;

    /** Constructor. */
    public DDTable() {
    	wDir = "";
    	wDirData = "data";
    }
    /** Base file name. */
	public String baseFile() {
		return wDir + fs + dataset;
	}
    /** ARFF file name. */
	public String arffFile() {
		return baseFile() + ".arff";
	}
    /** Decode items file name. */
	public String decodeFile() {
		return baseFile() + ".decode";
	}
    /** FIMI file name. */
	public String fimiFile() {
		return baseFile() + ".fimi";
	}
    /** Frequent itemsets file name. */
	public String pattFile() {
		return wDir + fs + wDirData + fs + dataset + "_" + support + ".patt";
	}

    /** Set working directory. */
	public void setWorkingDir(String wdir) {
		this.wDir = wdir;
	}
    /** Set working directory and data sub-directory. */
	public void setWorkingDir(String wdir, String data) {
		this.wDir = wdir;
		this.wDirData = data;
	}
    /** Load dataset from ARFF. */
    public void loadFromArff(String dataset) throws IOException {
	    loadFromArff(dataset, false);
    }
    /** Load dataset from ARFF. */
    public void loadFromArff(String dataset, boolean forceDecode) throws IOException {
	    if(dataset.endsWith(".arff"))
	    	this.dataset = dataset.substring(0, dataset.length()-5);
	    else
		    this.dataset = dataset;
	    File f = new File(decodeFile());
	    if(forceDecode || !f.exists()) { 
	    	// decode att=val into integers to the decodeFile()
	    	String cmd = arffFile() + " " + fimiFile() + " " + decodeFile();
	    	String cmdargs[] = cmd.split("\\s+");
	    	CSV2FIMI.main(cmdargs);
	    }
	    dec = new Decode(decodeFile());
	    nRows = IOUtil.countRows(fimiFile(), " ");
		table = new BitTable(dec.numberOfCodes(), nRows);
		table.loadFromFimi(fimiFile());
   }
   /** Number of attributes in the dataset. */
   public int getNoAtts() {
	   return dec.getNoAtts();
   }
   /** Decision attribute. */
   public int getDecAtt() {
	   return dec.getAttribute(decisionItem);
   }
  /** Set discrimination meta-data. */
   public void setDiscMetaData(String pd, String negDecision, int minSupp) {
	    support = minSupp;
	    decisionItem = dec.encode(negDecision);
	    int decisionItemCount = dec.support(decisionItem); // same as table.atts[decisionItem].bitCount();
	    double avgNegative =  decisionItemCount/(double)nRows;
	    pdItemString = pd;
	    decItemString = negDecision;
	    String []pds = pd.split(",");
	    pdItems = new ArrayList<Integer>();
	    for(String pdItem : pds) 
	    	pdItems.add( dec.encode(pdItem) );
	    SubsetGen<Integer> g = new SubsetGen<Integer>(pdItems);
		ArrayList<ArrayList<Integer>> subsets = g.getSubsets();
		specs = new CTSpecs[subsets.size()-1];
		int specpos = 0;
		for(ArrayList<Integer> set : subsets) {
			int ssize = set.size();
			if(ssize==0)
				continue;
		    Bitmap cover = table.atts[set.get(0)].clone();
		    for(int i=1; i<ssize; ++i)
		    	cover.and_inplace(table.atts[set.get(i)]);
		    int supp = cover.bitCount();
//			System.out.println(set+" "+supp);
		    double avgProtected =  supp/(double)nRows;
		    FrequentItemset pditemset = new FrequentItemset(dec.numberOfCodes(), supp);
		    for(Integer i : set)
		    	pditemset.set(i);
		    specs[specpos++]  = new CTSpecs(dec, pditemset, cover, avgProtected, avgNegative);
		}
	}
	/** Extract closed itemsets (not required if already on disk). */
	public void extractItemsets() throws IOException, InterruptedException
	{
		extractItemsets(null);
	}
	/** Extract closed itemsets (not required if extracted items already on disk). */
	public void extractItemsets(String excludeAtts) throws IOException, InterruptedException
	{
		int nc = dec.numberOfCodes();
		// remove decision attribute and protected items
		int decisionAtt = dec.getAttribute(decisionItem);
		TreeSet<Integer> excludeNo = new TreeSet<Integer>();
		excludeNo.add(decisionAtt);
		if(excludeAtts != null) 
			for(String s : excludeAtts.split(","))
				excludeNo.add( dec.getAttNo(s) );
		for(int i=0; i<nc; ++i) {
			int att = dec.getAttribute(i);
			dec.setRemoved(i, excludeNo.contains(att));
		}
		if(pdItems.size()==1) { // single PD item optimization
			int pdAtt = dec.getAttribute(pdItems.get(0));
			for(int i=0; i<nc; ++i) {
				int att = dec.getAttribute(i);
				if(att==pdAtt)
					dec.setRemoved(i, true);
			}
		}
		else
			for(Integer i : pdItems)
				dec.setRemoved(i, true);
		String temp = "tmp.pruned";
		dec.removeItems(fimiFile(), temp);
		// run FPGrowth
		String cmd = "lib/fpgrowth.exe " + (closedItemset ? "-tc" : "") + " -s-" + support + " -m0 " + temp + " " + pattFile();
//		System.out.println(cmd);
 	   	Runtime run = Runtime.getRuntime();
	    Process p = run.exec(cmd);
	    BufferedReader bis = new BufferedReader(new InputStreamReader(p.getErrorStream()));
	    String lines;
	    while( (lines = bis.readLine()) != null) {
	    	System.out.println(lines);
	    	System.out.flush();
	    }
	    bis.close();
	    p.waitFor();
	    File f = new File(temp);
	    f.delete();
	} 
	/** Get cover of a FrequentItemset. */
	public Bitmap getCover(FrequentItemset t) {
		ArrayList<Integer> items = t.getItems();
		if( items.size() == 0) {
			Bitmap res = new Bitmap(nRows);
			res.setall();
			return res;
		}
		Bitmap bm = null;
		boolean first = true;
		for(int i : items)
			if(first) {
				bm = table.atts[i].clone();
				first = false;
			} else
				bm.and_inplace(table.atts[i]);
		return bm;
	}
	/** Get cover of the decision item. */
	public Bitmap getCoverDecisionItem() {
		return table.atts[decisionItem];
	}
	/** Set values in a ContingencyTable for the whole table as context. */
	public void setCT(ContingencyTable ct) {
		int n = nRows;
		Bitmap bm1 = ct.specs.coverPD;;
		int n1 = bm1.bitCount();
		int m1 = dec.support(decisionItem); // same as table.atts[decisionItem].bitCount();
		ct.a = bm1.and_size(table.atts[decisionItem]);
		ct.b = n1 - ct.a;
		ct.c = m1 - ct.a;
		ct.d = n - n1 - ct.c;
		return;
	}
	/** Set values in a ContingencyTable for a given FrequentItemset as context. */
	public void setCT(ContingencyTable ct, FrequentItemset t) {
		ct.frset = t;
		int n = t.getSupport();
		ArrayList<Integer> items = t.getItems();
		if( items.size() == 0) {
			Bitmap bm1 = ct.specs.coverPD;;
			int n1 = bm1.bitCount();
			int m1 = dec.support(decisionItem); // same as table.atts[decisionItem].bitCount();
			ct.a = bm1.and_size(table.atts[decisionItem]);
			ct.b = n1 - ct.a;
			ct.c = m1 - ct.a;
			ct.d = n - n1 - ct.c;
			return;
		}
		Bitmap bm = null;
		boolean first = true;
		for(int i : items)
			if(first) {
				bm = table.atts[i].clone();
				first = false;
			} else
				bm.and_inplace(table.atts[i]);
		setCT(ct, bm, n);
	}
	/** Set values in a ContingencyTable for a given Bitmap of rows and support of the context. */
	public void setCT(ContingencyTable ct, Bitmap b, int n) {
		int m1 = b.and_size(table.atts[decisionItem]);
		Bitmap bm1 = b.and(ct.specs.coverPD);
		int n1 = bm1.bitCount();
		ct.a = bm1.and_size(table.atts[decisionItem]);
		ct.b = n1 - ct.a;
		ct.c = m1 - ct.a;
		ct.d = n - n1 - ct.c;	
	}
	/** Set values in a ContingencyTable for a given TableSubset. */
	public void setCT(ContingencyTable ct, TableSubset subset) {
		setCT(ct, subset.toBitmap(nRows), subset.size());
	}
	// SCAN OVER CONTINGENCY TABLES
	/** Current itemset read. */
	private FrequentItemset nextCTfrequent; 
	/** Which PD itemset is next. */
	private int nextCTstep;
	/** Start a scan of frequent (closed) itemsets. */
	public void initScan() throws IOException 
	{
	   // process closed itemsets
	   if(closedin != null)
		   closedin.close();
	   closedin = IOUtil.getReader(pattFile());
	   nextCTstep = 0;
	   nextCTfrequent = null;
	}
	/** Return a new object with the next contingency table, if any. */
	public ContingencyTable nextCT() throws Exception 
	{
		do {
		   if( nextCTstep == 0 ) {
		  	   String line;
		  	   if((line=closedin.readLine()) == null)
		  		   return null;
		  	   nextCTfrequent = FrequentItemset.parse(dec.numberOfCodes(), line);
			   nextCTstep = specs.length;
		   }
		   nextCTstep--;
		} while( specs.length > 1 && specs[nextCTstep].includeAttsFrom(nextCTfrequent) ); // context and pd items should not share any attribute
		// BUG: if context and pd items (>1) share an attribute & closed itemsets set!! To be corrected
	   ContingencyTable ct = new ContingencyTable(specs[nextCTstep]);
	   setCT(ct, nextCTfrequent);
	   return ct;
	}
	/** End action of a scan. */
	public void endScan() throws IOException 
	{
	   // closed itemsets
	   if(closedin != null)
		   closedin.close();
	}

	// METHODS USEFUL FOR FRECHET ATTACKS
	/** Returns items not involved in a contingency table. */
	public int[] getItemsNotIn(ContingencyTable ct) {
		ArrayList<Integer> items = ct.frset.getItems();
		// compute attributes of items in the frequent itemset
		TreeSet<Integer> atts = new TreeSet<Integer>();
		for(int item : items)
			atts.add( dec.getAttribute(item) );
		atts.add( dec.getAttribute(decisionItem) );
		// calculate size of returned array
		int nres = 0;
		int nitems = dec.numberOfCodes();
		for(int i=0; i<nitems; ++i)
			if( !atts.contains(dec.getAttribute(i)) )
				nres++;
		// create returned array
		int []res = new int[nres];
		nres = 0;
		for(int i=0; i<nitems; ++i)
			if( !atts.contains(dec.getAttribute(i)) )
				res[nres++] = i;
		return res;
	}
	/** Return a ContingencyTable obtained by changing the PD item. Warning: CTSpecs are not set.*/
	public ContingencyTable changePD(ContingencyTable ct, int rowItem) {
		ContingencyTable res = new ContingencyTable();
		res.frset = ct.frset;
		int n = ct.n();
		int m1 = ct.m1();
		ArrayList<Integer> items = ct.frset.getItems();
		Bitmap bm1 = null;
		if( items.size() == 0) {
			bm1 = table.atts[rowItem];
		} else {
			boolean first = true;
			for(int i : items)
				if(first) {
					bm1 = table.atts[i].clone();
					first = false;
				} else
					bm1.and_inplace(table.atts[i]);
			bm1.and_inplace(table.atts[rowItem]);
		}
		int n1 = bm1.bitCount();
		res.a = bm1.and_size(table.atts[decisionItem]);
		res.b = n1 - res.a;
		res.c = m1 - res.a;
		res.d = n - n1 - res.c;	
		return res;
	}
	/** Return a ContingencyTable obtained by changing the decision item. Warning: CTSpecs are not set.*/
	public ContingencyTable changeDec(ContingencyTable ct, int decItem) {
		ContingencyTable res = new ContingencyTable();
		res.frset = ct.frset;
		int n = ct.n();
		int n1 = ct.n1();
		ArrayList<Integer> items = ct.frset.getItems();
		Bitmap bm1 = null;
		if( items.size() == 0) {
			bm1 = table.atts[decItem];
		} else {
			boolean first = true;
			for(int i : items)
				if(first) {
					bm1 = table.atts[i].clone();
					first = false;
				} else
					bm1.and_inplace(table.atts[i]);
			bm1.and_inplace(table.atts[decItem]);
		}
		int m1 = bm1.bitCount();
		res.a = bm1.and_size(ct.specs.coverPD);
		res.b = n1 - res.a;
		res.c = m1 - res.a;
		res.d = n - n1 - res.c;	
		return res;
	}
	
	// METHODS USEFUL FOR DANONYMITY
	/** Ranges for t values. */
	public double tValueMin, tValueMax, avgNegProMin;
	/** tvalue range over all PD itemsets. */
	public void setTRanges(boolean sdlike) {
		tValueMin = 0;
		tValueMax = 1;
		avgNegProMin = 1;
		for(CTSpecs spec : specs) {
			ContingencyTable ct = new ContingencyTable(spec);
			setCT(ct);
			double mint = ct.tval(sdlike); 
			if( mint > tValueMin )
				tValueMin = mint;
			double negPro = sdlike ? ct.specs.avgProtected : ct.specs.avgNegative;
			if(negPro<avgNegProMin)
				avgNegProMin = negPro;
			double maxt = sdlike ? Math.max(ct.specs.avgProtected, 1 - ct.specs.avgProtected) :
							Math.max(ct.specs.avgNegative, 1 - ct.specs.avgNegative);
			if(maxt < tValueMax)
				tValueMax = maxt;
		}
	}
	/** Return bound on alpha-discrimination for a dataset sanitized for t-closeness. */
	public double getBound(Measure m, double t) {
		switch(m) {
		case SD:
		case RD: 
		case ED: return Math.min(1,Math.min(2*t, t+(1-tValueMax)));
		case RR: 
		case ER: return Math.min(1,avgNegProMin+t)/Math.max(0,avgNegProMin-t);
		case EC: 
		case RC: return Math.max(0, 1-avgNegProMin-t)/Math.min(1,1-avgNegProMin+t);
		case OR: return Math.min(1,avgNegProMin+t)*Math.min(1,1-avgNegProMin+t)/
							(Math.max(0,avgNegProMin-t)*Math.max(0,1-avgNegProMin-t));
		default: throw new RuntimeException("Measure " + m + " unknown");
		}
	}
	/** Inverse bound function. */
	public double getBoundInverse(Measure m, double alpha) {
		switch(m) {
		case SD:
		case RD: 
		case ED: return Math.max(alpha/2, alpha-(1-tValueMax));
		case RR: 
		case ER: return avgNegProMin*(alpha-1)/(alpha+1);
		case EC: 
		case RC: return (1-avgNegProMin)*(1-alpha)/(1+alpha);
		case OR: 
			double a = 1-alpha;
			double b = 1+alpha;
			double c = avgNegProMin*(1-avgNegProMin)*(1-alpha);
			return (-b + Math.sqrt(b*b-4*a*c))/(2*a);
		default: throw new RuntimeException("Measure " + m + " unknown");
		}
	}
}
