package ags.utils.dataStructures.trees.thirdGenKD;

import java.util.Arrays;
import java.util.TreeMap;

/**
 *
 */
class KdNode<T> implements Comparable<KdNode<T>>{
    // All types
    protected int dimensions;
    protected int bucketCapacity;
    protected int size;

    // Leaf only
    protected double[][] points;
    protected Object[] data;

    // Stem only
    protected KdNode<T> left, right;
    protected int splitDimension;
    protected double splitValue;

    // Bounds
    protected double[] minBound, maxBound;
    protected boolean singlePoint;

    protected KdNode(int dimensions, int bucketCapacity) {
        // Init base
        this.dimensions = dimensions;
        this.bucketCapacity = bucketCapacity;
        this.size = 0;
        this.singlePoint = true;

        // Init leaf elements
        this.points = new double[bucketCapacity+1][];
        this.data = new Object[bucketCapacity+1];
    }

    /* -------- SIMPLE GETTERS -------- */

    public int size() {
        return size;
    }

    public boolean isLeaf() {
        return points != null;
    }

    /* -------- OPERATIONS -------- */

    public void addPoint(double[] point, T value) {
        KdNode<T> cursor = this;
        while (!cursor.isLeaf()) {
            cursor.extendBounds(point);
            cursor.size++;
            if (point[cursor.splitDimension] > cursor.splitValue) {
                cursor = cursor.right;
            }
            else {
                cursor = cursor.left;
            }
        }
        cursor.addLeafPoint(point, value);
    }

    /* -------- DELETE OPERATIONS added by Salvatore Ruggieri (c) 2013 -------- */

 	public T getRandomPoint() {
    	return hashmap.firstKey();
    }
    
    public void deletePoint(T value) {
    	KdNode<T> node = hashmap.get(value);
    	node._deletePoint(value);
    	hashmap.remove(value);
    	if(node != this)
    		--size; // size in between the root and the leaf become inconsistent
    }
    
	private void _deletePoint(T value) {
    	for(int i=0; i<size; ++i)
    		if(value.equals(data[i])) {
/*        		double []p = points[i];
        		int n = p.length;
        		double factor = root.npoints/(root.npoints-1);
        		for(int j=0;j<n;++j)
        			root.centroid[j] = (root.centroid[j] - p[j]/root.npoints)*factor;
        		root.npoints--;
*/
    			for(int j=i+1; j<size; ++j) {
	    			points[j-1] = points[j];
	    			data[j-1] = data[j];
    			}
    		--size;
    		return;
    		}
    	throw new RuntimeException("_deletePoint inconsistent");
    }
 
    TreeMap<T, KdNode<T>> hashmap;
/*    double []centroid;
    double npoints;
    
    public double[] getCentroid() {
    	return centroid;
    }
*/    
    public void prepareDeletePoints() {
    	hashmap = new TreeMap<T, KdNode<T>>();
//   	centroid = null;
//    	npoints = this.size();
    	_prepareDeletePoints(hashmap);
    }
    
    @SuppressWarnings("unchecked")
	public void _prepareDeletePoints(TreeMap<T, KdNode<T>> hashmap) {
    	if(isLeaf()) {
        	for(int i=0; i<size; ++i) {
        		T v = (T)(data[i]);
        		hashmap.put(v, this);
/*        		double []p = points[i];
        		int n = p.length;
        		if(root.centroid==null)
        			root.centroid = new double[n];
        		for(int j=0;j<n;++j)
        			root.centroid[j] += p[j]/npoints;
*/
        	}
        	return;
    	}
    	left._prepareDeletePoints(hashmap);
    	right._prepareDeletePoints(hashmap);
    }
     
	@Override
	public int compareTo(KdNode<T> o) {
		if( this == o)
			return 0;
		return -1;
	}

	/* -------- INTERNAL OPERATIONS -------- */

    public void addLeafPoint(double[] point, T value) {
        // Add the data point
        points[size] = point;
        data[size] = value;
        extendBounds(point);
        size++;

        if (size == points.length - 1) {
            // If the node is getting too large
            if (calculateSplit()) {
                // If the node successfully had it's split value calculated, split node
                splitLeafNode();
            } else {
                // If the node could not be split, enlarge node
                increaseLeafCapacity();
            }
        }
    }

    @SuppressWarnings("unused")
	private boolean checkBounds(double[] point) {
        for (int i = 0; i < dimensions; i++) {
            if (point[i] > maxBound[i]) return false;
            if (point[i] < minBound[i]) return false;
        }
        return true;
    }

    private void extendBounds(double[] point) {
        if (minBound == null) {
            minBound = Arrays.copyOf(point, dimensions);
            maxBound = Arrays.copyOf(point, dimensions);
            return;
        }

        for (int i = 0; i < dimensions; i++) {
            if (Double.isNaN(point[i])) {
                if (!Double.isNaN(minBound[i]) || !Double.isNaN(maxBound[i])) {
                    singlePoint = false;
                }
                minBound[i] = Double.NaN;
                maxBound[i] = Double.NaN;
            }
            else if (minBound[i] > point[i]) {
                minBound[i] = point[i];
                singlePoint = false;
            }
            else if (maxBound[i] < point[i]) {
                maxBound[i] = point[i];
                singlePoint = false;
            }
        }
    }

    private void increaseLeafCapacity() {
        points = Arrays.copyOf(points, points.length*2);
        data = Arrays.copyOf(data, data.length*2);
    }

    private boolean calculateSplit() {
        if (singlePoint) return false;

        double width = 0;
        for (int i = 0; i < dimensions; i++) {
            double dwidth = (maxBound[i] - minBound[i]);
            if (Double.isNaN(dwidth)) dwidth = 0;
            if (dwidth > width) {
                splitDimension = i;
                width = dwidth;
            }
        }

        if (width == 0) {
            return false;
        }

        // Start the split in the middle of the variance
        splitValue = (minBound[splitDimension] + maxBound[splitDimension]) * 0.5;

        // Never split on infinity or NaN
        if (splitValue == Double.POSITIVE_INFINITY) {
            splitValue = Double.MAX_VALUE;
        }
        else if (splitValue == Double.NEGATIVE_INFINITY) {
            splitValue = -Double.MAX_VALUE;
        }

        // Don't let the split value be the same as the upper value as
        // can happen due to rounding errors!
        if (splitValue == maxBound[splitDimension]) {
            splitValue = minBound[splitDimension];
        }

        // Success
        return true;
    }

    @SuppressWarnings("unchecked")
	private void splitLeafNode() {
        right = new KdNode<T>(dimensions, bucketCapacity);
        left = new KdNode<T>(dimensions, bucketCapacity);

        // Move locations into children
        for (int i = 0; i < size; i++) {
            double[] oldLocation = points[i];
            Object oldData = data[i];
            if (oldLocation[splitDimension] > splitValue) {
                right.addLeafPoint(oldLocation, (T) oldData);
            }
            else {
                left.addLeafPoint(oldLocation, (T) oldData);
            }
        }

        points = null;
        data = null;
    }
}
