/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-
 *    Salvatore Ruggieri
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package rst.pp.fimi;

import java.io.*;
import java.util.*;

import rst.util.IOUtil;

/**
 * CSV2FIMI: command line program for producing a FIMI file format 
 * 			 starting from a CSV file format with header or from a ARFF file format.
 */
public class CSV2FIMI {

	public static ColumnCode[] csv2fimi_pass(BufferedReader in, ColumnCode []ctypein,  boolean is_last) 
	throws IOException 
	{
		ArrayList<String> atts = getCSVattributes(in);
	    int ncols = atts.size();
	    ColumnCode[] ctype = (ctypein != null) ? ctypein : new ColumnCode[ncols];
	    // create columns
	    if( ctypein == null )
		    for (int i = 0; i < ncols; ++i) 
		        ctype[i] = new ColumnCode(atts.get(i));
	    // pass on tokens
	    String[] tokens;
		while ((tokens = IOUtil.readTokens(in)) != null) {
			if( tokens.length == 0 )
				continue;
		    for (int i = 0; i < ncols; ++i) 
		        ctype[i].addCode(tokens[i]);
		}
		// last file - set base for columns
		if( is_last ) {
			int sum = 0;
		    for (int i = 0; i < ncols-1; ++i) {
		    	sum += ctype[i].sequence-1;
		    	ctype[i+1].setBase( sum );
		    }
		}
		return ctype;
	}
	
	/** FIMI output. */
	public static void csv2fimi(ColumnCode[] ctype, BufferedReader in, PrintWriter out) 
	throws IOException 
	{	
		getCSVattributes(in);
	    int ncols = ctype.length;
	    String[] tokens;
		while ((tokens = IOUtil.readTokens(in)) != null) {
			if( tokens.length == 0 )
				continue;
		    for (int i = 0; i < ncols-1; ++i) 
		        out.print( ctype[i].getCode(tokens[i]) + " " );
	        out.println( ctype[ncols-1].getCode(tokens[ncols-1]) );
		}
	}
	
	/** Read attribute names from ARFF header or CSV header. */
	public static ArrayList<String> getCSVattributes(BufferedReader in) 
	throws IOException 
	{
		ArrayList<String> res = new ArrayList<String>();
		String line = in.readLine();
		if (line.toLowerCase().startsWith("@relation")) 
		{ // ARFF file
			while ((line = in.readLine()) != null) {
				if (line.toLowerCase().startsWith("@data"))
					break;
				if (line.toLowerCase().startsWith("@attribute")) {
					StringTokenizer tk = new StringTokenizer(line, " ");
					tk.nextToken();
					String attname = tk.nextToken();
					res.add(attname);
				}
			}
		} else {
			// CSV file with header
	        StringTokenizer st = new StringTokenizer(line, ",");
	        while (st.hasMoreTokens())
	            res.add( st.nextToken() );
		}
		return res;
	}
	
	public static void main(String[] args) 
	{
	    if ( (args.length < 3) || (args.length % 2 == 0) ) {
	        System.out.println("Usage: java CSV2FIMI csv_in+ fimi_out+ meta_out");
	            System.exit(0);
	        }
	    	int nfiles = (args.length - 1) / 2;
	        try {
	        	ColumnCode[] atts = null;
	        	// first pass
	        	for(int i=0; i<nfiles; ++i) {
	        		BufferedReader input = IOUtil.getReader(args[i]);
	        		atts = csv2fimi_pass(input, atts, (i==nfiles-1));
	        		input.close();
	        	}
	        	// second pass
	        	for(int i=0; i<nfiles; ++i) {
	        		BufferedReader input = IOUtil.getReader(args[i]);
	        		PrintWriter output = IOUtil.getWriter(args[i+nfiles]);
		            csv2fimi(atts, input,output);
		            input.close();
		            output.close();
	        	}
	        	// output codes
	        	PrintWriter decode = IOUtil.getWriter(args[args.length-1]);
	    	    for (int i = 0; i < atts.length; ++i)
	    	    	atts[i].output(decode);
	            decode.close();
	           } catch (IOException e) {
	            System.out.println(e.toString());
	        }
	    }
}
